!-------------------------------------------------------------------------------
! (C) British Crown Copyright 2020 Met Office
!
! This software is licensed under the terms of the Apache Licence Version 2.0
! which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
!-------------------------------------------------------------------------------

!> Fortran module with various useful routines

module ufo_utils_mod

use fckit_log_module, only : fckit_log
use kinds, only: kind_real
use missing_values_mod, only: missing_value
use ufo_constants_mod, only : zero, half, one, two, &
                              t0c, zerodegc, min_q, epsilon

implicit none
private

public Ops_Qsat
public Ops_QsatWat
public Ops_SatRad_Qsplit
public Ops_Cholesky
public ufo_utils_iogetfreeunit
public InvertMatrix
public upper2lower
public getindex
public cmp_strings
public find_unique
public Ops_RealSortQuick
public sort_and_unique

contains

!-------------------------------------------------------------------------------
!> Calculate the Saturation Specific Humidity Scheme (Qsat): Vapour to Liquid/Ice.
!!
!! \details Heritage: Ops_Qsat.inc
!!
!! Returns a saturation mixing ratio given a temperature and pressure
!! using saturation vapour pressures caluclated using the Goff-Gratch
!! formulae, adopted by the WMO as taken from Landolt-Bornstein, 1987
!! Numerical data and Functional relationships in Science and
!! Technology.  Group V/Vol 4B Meteorology.  Physical and Chemical
!! properties of Air, P35.
!!
!! Value in the lookup table are over water above 0 degrees C and over
!! ice below this temperatures.
!!
!! Method: <br>
!!   uses lookup tables to find eSAT, calculates qSAT directly from that.
!!
!! \author Met Office
!!
!! \date 09/06/2020: Created
!!
subroutine Ops_Qsat (QS, &
                      T, &
                      P, &
                     npnts)

implicit none

! subroutine arguments:
integer, intent(in)                 :: npnts     !< Points being processed by qSAT scheme.
real(kind=kind_real), intent(in)    :: T(npnts)  !< Temperature (K)
real(kind=kind_real), intent(in)    :: P(npnts)  !< Pressure (Pa).
real(kind=kind_real), intent(inout) :: QS(npnts) !< Saturation mixing ratio (KG/KG)

! Local declarations:
real(kind=kind_real), parameter :: one_minus_epsilon = one - epsilon
real(kind=kind_real), parameter :: T_low = 183.15_kind_real  ! Lowest temperature for which look-up table is valid
real(kind=kind_real), parameter :: T_high = 338.15_kind_real  ! Highest temperature for which look-up table is valid
real(kind=kind_real), parameter :: delta_T = 0.1_kind_real    ! Temperature increment of look-up table
integer, parameter   :: N = ((T_high - T_low + (delta_T * half)) / delta_T) + one ! Size of lookup-table (gives 1551)
integer              :: ITABLE
real(kind=kind_real) :: ATABLE
real(kind=kind_real) :: FSUBW      ! Converts from sat vapour pressure in pure water to pressure in air
real(kind=kind_real) :: TT
integer              :: I
integer              :: IES
real(kind=kind_real) :: ES(0:N + 1)    ! Table of saturation water vapour pressure (PA)
character(len=*), parameter :: RoutineName = "Ops_Qsat"

! Note: 0 element is a repeat of 1st element to cater for special case
!       of low temperatures (.LE.T_low) for which the array index is
!       rounded down due to machine precision.

data (ES(IES), IES= 0, 95) / 0.966483D-02, &
0.966483D-02,0.984279D-02,0.100240D-01,0.102082D-01,0.103957D-01, &
0.105865D-01,0.107803D-01,0.109777D-01,0.111784D-01,0.113825D-01, &
0.115902D-01,0.118016D-01,0.120164D-01,0.122348D-01,0.124572D-01, &
0.126831D-01,0.129132D-01,0.131470D-01,0.133846D-01,0.136264D-01, &
0.138724D-01,0.141225D-01,0.143771D-01,0.146356D-01,0.148985D-01, &
0.151661D-01,0.154379D-01,0.157145D-01,0.159958D-01,0.162817D-01, &
0.165725D-01,0.168680D-01,0.171684D-01,0.174742D-01,0.177847D-01, &
0.181008D-01,0.184216D-01,0.187481D-01,0.190801D-01,0.194175D-01, &
0.197608D-01,0.201094D-01,0.204637D-01,0.208242D-01,0.211906D-01, &
0.215631D-01,0.219416D-01,0.223263D-01,0.227172D-01,0.231146D-01, &
0.235188D-01,0.239296D-01,0.243465D-01,0.247708D-01,0.252019D-01, &
0.256405D-01,0.260857D-01,0.265385D-01,0.269979D-01,0.274656D-01, &
0.279405D-01,0.284232D-01,0.289142D-01,0.294124D-01,0.299192D-01, &
0.304341D-01,0.309571D-01,0.314886D-01,0.320285D-01,0.325769D-01, &
0.331348D-01,0.337014D-01,0.342771D-01,0.348618D-01,0.354557D-01, &
0.360598D-01,0.366727D-01,0.372958D-01,0.379289D-01,0.385717D-01, &
0.392248D-01,0.398889D-01,0.405633D-01,0.412474D-01,0.419430D-01, &
0.426505D-01,0.433678D-01,0.440974D-01,0.448374D-01,0.455896D-01, &
0.463545D-01,0.471303D-01,0.479191D-01,0.487190D-01,0.495322D-01/
      data (ES(IES),IES= 96,190) / &
0.503591D-01,0.511977D-01,0.520490D-01,0.529145D-01,0.537931D-01, &
0.546854D-01,0.555924D-01,0.565119D-01,0.574467D-01,0.583959D-01, &
0.593592D-01,0.603387D-01,0.613316D-01,0.623409D-01,0.633655D-01, &
0.644053D-01,0.654624D-01,0.665358D-01,0.676233D-01,0.687302D-01, &
0.698524D-01,0.709929D-01,0.721490D-01,0.733238D-01,0.745180D-01, &
0.757281D-01,0.769578D-01,0.782061D-01,0.794728D-01,0.807583D-01, &
0.820647D-01,0.833905D-01,0.847358D-01,0.861028D-01,0.874882D-01, &
0.888957D-01,0.903243D-01,0.917736D-01,0.932464D-01,0.947407D-01, &
0.962571D-01,0.977955D-01,0.993584D-01,0.100942D+00,0.102551D+00, &
0.104186D+00,0.105842D+00,0.107524D+00,0.109231D+00,0.110963D+00, &
0.112722D+00,0.114506D+00,0.116317D+00,0.118153D+00,0.120019D+00, &
0.121911D+00,0.123831D+00,0.125778D+00,0.127755D+00,0.129761D+00, &
0.131796D+00,0.133863D+00,0.135956D+00,0.138082D+00,0.140241D+00, &
0.142428D+00,0.144649D+00,0.146902D+00,0.149190D+00,0.151506D+00, &
0.153859D+00,0.156245D+00,0.158669D+00,0.161126D+00,0.163618D+00, &
0.166145D+00,0.168711D+00,0.171313D+00,0.173951D+00,0.176626D+00, &
0.179342D+00,0.182096D+00,0.184893D+00,0.187724D+00,0.190600D+00, &
0.193518D+00,0.196473D+00,0.199474D+00,0.202516D+00,0.205604D+00, &
0.208730D+00,0.211905D+00,0.215127D+00,0.218389D+00,0.221701D+00/
      data (ES(IES),IES=191,285) / &
0.225063D+00,0.228466D+00,0.231920D+00,0.235421D+00,0.238976D+00, &
0.242580D+00,0.246232D+00,0.249933D+00,0.253691D+00,0.257499D+00, &
0.261359D+00,0.265278D+00,0.269249D+00,0.273274D+00,0.277358D+00, &
0.281498D+00,0.285694D+00,0.289952D+00,0.294268D+00,0.298641D+00, &
0.303078D+00,0.307577D+00,0.312135D+00,0.316753D+00,0.321440D+00, &
0.326196D+00,0.331009D+00,0.335893D+00,0.340842D+00,0.345863D+00, &
0.350951D+00,0.356106D+00,0.361337D+00,0.366636D+00,0.372006D+00, &
0.377447D+00,0.382966D+00,0.388567D+00,0.394233D+00,0.399981D+00, &
0.405806D+00,0.411714D+00,0.417699D+00,0.423772D+00,0.429914D+00, &
0.436145D+00,0.442468D+00,0.448862D+00,0.455359D+00,0.461930D+00, &
0.468596D+00,0.475348D+00,0.482186D+00,0.489124D+00,0.496160D+00, &
0.503278D+00,0.510497D+00,0.517808D+00,0.525224D+00,0.532737D+00, &
0.540355D+00,0.548059D+00,0.555886D+00,0.563797D+00,0.571825D+00, &
0.579952D+00,0.588198D+00,0.596545D+00,0.605000D+00,0.613572D+00, &
0.622255D+00,0.631059D+00,0.639962D+00,0.649003D+00,0.658144D+00, &
0.667414D+00,0.676815D+00,0.686317D+00,0.695956D+00,0.705728D+00, &
0.715622D+00,0.725641D+00,0.735799D+00,0.746082D+00,0.756495D+00, &
0.767052D+00,0.777741D+00,0.788576D+00,0.799549D+00,0.810656D+00, &
0.821914D+00,0.833314D+00,0.844854D+00,0.856555D+00,0.868415D+00/
      data (ES(IES),IES=286,380) / &
0.880404D+00,0.892575D+00,0.904877D+00,0.917350D+00,0.929974D+00, &
0.942771D+00,0.955724D+00,0.968837D+00,0.982127D+00,0.995600D+00, &
0.100921D+01,0.102304D+01,0.103700D+01,0.105116D+01,0.106549D+01, &
0.108002D+01,0.109471D+01,0.110962D+01,0.112469D+01,0.113995D+01, &
0.115542D+01,0.117107D+01,0.118693D+01,0.120298D+01,0.121923D+01, &
0.123569D+01,0.125234D+01,0.126923D+01,0.128631D+01,0.130362D+01, &
0.132114D+01,0.133887D+01,0.135683D+01,0.137500D+01,0.139342D+01, &
0.141205D+01,0.143091D+01,0.145000D+01,0.146933D+01,0.148892D+01, &
0.150874D+01,0.152881D+01,0.154912D+01,0.156970D+01,0.159049D+01, &
0.161159D+01,0.163293D+01,0.165452D+01,0.167640D+01,0.169852D+01, &
0.172091D+01,0.174359D+01,0.176653D+01,0.178977D+01,0.181332D+01, &
0.183709D+01,0.186119D+01,0.188559D+01,0.191028D+01,0.193524D+01, &
0.196054D+01,0.198616D+01,0.201208D+01,0.203829D+01,0.206485D+01, &
0.209170D+01,0.211885D+01,0.214637D+01,0.217424D+01,0.220242D+01, &
0.223092D+01,0.225979D+01,0.228899D+01,0.231855D+01,0.234845D+01, &
0.237874D+01,0.240937D+01,0.244040D+01,0.247176D+01,0.250349D+01, &
0.253560D+01,0.256814D+01,0.260099D+01,0.263431D+01,0.266800D+01, &
0.270207D+01,0.273656D+01,0.277145D+01,0.280671D+01,0.284248D+01, &
0.287859D+01,0.291516D+01,0.295219D+01,0.298962D+01,0.302746D+01/
      data (ES(IES),IES=381,475) / &
0.306579D+01,0.310454D+01,0.314377D+01,0.318351D+01,0.322360D+01, &
0.326427D+01,0.330538D+01,0.334694D+01,0.338894D+01,0.343155D+01, &
0.347456D+01,0.351809D+01,0.356216D+01,0.360673D+01,0.365184D+01, &
0.369744D+01,0.374352D+01,0.379018D+01,0.383743D+01,0.388518D+01, &
0.393344D+01,0.398230D+01,0.403177D+01,0.408175D+01,0.413229D+01, &
0.418343D+01,0.423514D+01,0.428746D+01,0.434034D+01,0.439389D+01, &
0.444808D+01,0.450276D+01,0.455820D+01,0.461423D+01,0.467084D+01, &
0.472816D+01,0.478607D+01,0.484468D+01,0.490393D+01,0.496389D+01, &
0.502446D+01,0.508580D+01,0.514776D+01,0.521047D+01,0.527385D+01, &
0.533798D+01,0.540279D+01,0.546838D+01,0.553466D+01,0.560173D+01, &
0.566949D+01,0.573807D+01,0.580750D+01,0.587749D+01,0.594846D+01, &
0.602017D+01,0.609260D+01,0.616591D+01,0.623995D+01,0.631490D+01, &
0.639061D+01,0.646723D+01,0.654477D+01,0.662293D+01,0.670220D+01, &
0.678227D+01,0.686313D+01,0.694495D+01,0.702777D+01,0.711142D+01, &
0.719592D+01,0.728140D+01,0.736790D+01,0.745527D+01,0.754352D+01, &
0.763298D+01,0.772316D+01,0.781442D+01,0.790676D+01,0.800001D+01, &
0.809435D+01,0.818967D+01,0.828606D+01,0.838343D+01,0.848194D+01, &
0.858144D+01,0.868207D+01,0.878392D+01,0.888673D+01,0.899060D+01, &
0.909567D+01,0.920172D+01,0.930909D+01,0.941765D+01,0.952730D+01/
      data (ES(IES),IES=476,570) / &
0.963821D+01,0.975022D+01,0.986352D+01,0.997793D+01,0.100937D+02, &
0.102105D+02,0.103287D+02,0.104481D+02,0.105688D+02,0.106909D+02, &
0.108143D+02,0.109387D+02,0.110647D+02,0.111921D+02,0.113207D+02, &
0.114508D+02,0.115821D+02,0.117149D+02,0.118490D+02,0.119847D+02, &
0.121216D+02,0.122601D+02,0.124002D+02,0.125416D+02,0.126846D+02, &
0.128290D+02,0.129747D+02,0.131224D+02,0.132712D+02,0.134220D+02, &
0.135742D+02,0.137278D+02,0.138831D+02,0.140403D+02,0.141989D+02, &
0.143589D+02,0.145211D+02,0.146845D+02,0.148501D+02,0.150172D+02, &
0.151858D+02,0.153564D+02,0.155288D+02,0.157029D+02,0.158786D+02, &
0.160562D+02,0.162358D+02,0.164174D+02,0.166004D+02,0.167858D+02, &
0.169728D+02,0.171620D+02,0.173528D+02,0.175455D+02,0.177406D+02, &
0.179372D+02,0.181363D+02,0.183372D+02,0.185400D+02,0.187453D+02, &
0.189523D+02,0.191613D+02,0.193728D+02,0.195866D+02,0.198024D+02, &
0.200200D+02,0.202401D+02,0.204626D+02,0.206871D+02,0.209140D+02, &
0.211430D+02,0.213744D+02,0.216085D+02,0.218446D+02,0.220828D+02, &
0.223241D+02,0.225671D+02,0.228132D+02,0.230615D+02,0.233120D+02, &
0.235651D+02,0.238211D+02,0.240794D+02,0.243404D+02,0.246042D+02, &
0.248704D+02,0.251390D+02,0.254109D+02,0.256847D+02,0.259620D+02, &
0.262418D+02,0.265240D+02,0.268092D+02,0.270975D+02,0.273883D+02/
      data (ES(IES),IES=571,665) / &
0.276822D+02,0.279792D+02,0.282789D+02,0.285812D+02,0.288867D+02, &
0.291954D+02,0.295075D+02,0.298222D+02,0.301398D+02,0.304606D+02, &
0.307848D+02,0.311119D+02,0.314424D+02,0.317763D+02,0.321133D+02, &
0.324536D+02,0.327971D+02,0.331440D+02,0.334940D+02,0.338475D+02, &
0.342050D+02,0.345654D+02,0.349295D+02,0.352975D+02,0.356687D+02, &
0.360430D+02,0.364221D+02,0.368042D+02,0.371896D+02,0.375790D+02, &
0.379725D+02,0.383692D+02,0.387702D+02,0.391744D+02,0.395839D+02, &
0.399958D+02,0.404118D+02,0.408325D+02,0.412574D+02,0.416858D+02, &
0.421188D+02,0.425551D+02,0.429962D+02,0.434407D+02,0.438910D+02, &
0.443439D+02,0.448024D+02,0.452648D+02,0.457308D+02,0.462018D+02, &
0.466775D+02,0.471582D+02,0.476428D+02,0.481313D+02,0.486249D+02, &
0.491235D+02,0.496272D+02,0.501349D+02,0.506479D+02,0.511652D+02, &
0.516876D+02,0.522142D+02,0.527474D+02,0.532836D+02,0.538266D+02, &
0.543737D+02,0.549254D+02,0.554839D+02,0.560456D+02,0.566142D+02, &
0.571872D+02,0.577662D+02,0.583498D+02,0.589392D+02,0.595347D+02, &
0.601346D+02,0.607410D+02,0.613519D+02,0.619689D+02,0.625922D+02, &
0.632204D+02,0.638550D+02,0.644959D+02,0.651418D+02,0.657942D+02, &
0.664516D+02,0.671158D+02,0.677864D+02,0.684624D+02,0.691451D+02, &
0.698345D+02,0.705293D+02,0.712312D+02,0.719398D+02,0.726542D+02/
      data (ES(IES),IES=666,760) / &
0.733754D+02,0.741022D+02,0.748363D+02,0.755777D+02,0.763247D+02, &
0.770791D+02,0.778394D+02,0.786088D+02,0.793824D+02,0.801653D+02, &
0.809542D+02,0.817509D+02,0.825536D+02,0.833643D+02,0.841828D+02, &
0.850076D+02,0.858405D+02,0.866797D+02,0.875289D+02,0.883827D+02, &
0.892467D+02,0.901172D+02,0.909962D+02,0.918818D+02,0.927760D+02, &
0.936790D+02,0.945887D+02,0.955071D+02,0.964346D+02,0.973689D+02, &
0.983123D+02,0.992648D+02,0.100224D+03,0.101193D+03,0.102169D+03, &
0.103155D+03,0.104150D+03,0.105152D+03,0.106164D+03,0.107186D+03, &
0.108217D+03,0.109256D+03,0.110303D+03,0.111362D+03,0.112429D+03, &
0.113503D+03,0.114588D+03,0.115684D+03,0.116789D+03,0.117903D+03, &
0.119028D+03,0.120160D+03,0.121306D+03,0.122460D+03,0.123623D+03, &
0.124796D+03,0.125981D+03,0.127174D+03,0.128381D+03,0.129594D+03, &
0.130822D+03,0.132058D+03,0.133306D+03,0.134563D+03,0.135828D+03, &
0.137109D+03,0.138402D+03,0.139700D+03,0.141017D+03,0.142338D+03, &
0.143676D+03,0.145025D+03,0.146382D+03,0.147753D+03,0.149133D+03, &
0.150529D+03,0.151935D+03,0.153351D+03,0.154783D+03,0.156222D+03, &
0.157678D+03,0.159148D+03,0.160624D+03,0.162117D+03,0.163621D+03, &
0.165142D+03,0.166674D+03,0.168212D+03,0.169772D+03,0.171340D+03, &
0.172921D+03,0.174522D+03,0.176129D+03,0.177755D+03,0.179388D+03/
      data (ES(IES),IES=761,855) / &
0.181040D+03,0.182707D+03,0.184382D+03,0.186076D+03,0.187782D+03, &
0.189503D+03,0.191240D+03,0.192989D+03,0.194758D+03,0.196535D+03, &
0.198332D+03,0.200141D+03,0.201963D+03,0.203805D+03,0.205656D+03, &
0.207532D+03,0.209416D+03,0.211317D+03,0.213236D+03,0.215167D+03, &
0.217121D+03,0.219087D+03,0.221067D+03,0.223064D+03,0.225080D+03, &
0.227113D+03,0.229160D+03,0.231221D+03,0.233305D+03,0.235403D+03, &
0.237520D+03,0.239655D+03,0.241805D+03,0.243979D+03,0.246163D+03, &
0.248365D+03,0.250593D+03,0.252830D+03,0.255093D+03,0.257364D+03, &
0.259667D+03,0.261979D+03,0.264312D+03,0.266666D+03,0.269034D+03, &
0.271430D+03,0.273841D+03,0.276268D+03,0.278722D+03,0.281185D+03, &
0.283677D+03,0.286190D+03,0.288714D+03,0.291266D+03,0.293834D+03, &
0.296431D+03,0.299045D+03,0.301676D+03,0.304329D+03,0.307006D+03, &
0.309706D+03,0.312423D+03,0.315165D+03,0.317930D+03,0.320705D+03, &
0.323519D+03,0.326350D+03,0.329199D+03,0.332073D+03,0.334973D+03, &
0.337897D+03,0.340839D+03,0.343800D+03,0.346794D+03,0.349806D+03, &
0.352845D+03,0.355918D+03,0.358994D+03,0.362112D+03,0.365242D+03, &
0.368407D+03,0.371599D+03,0.374802D+03,0.378042D+03,0.381293D+03, &
0.384588D+03,0.387904D+03,0.391239D+03,0.394604D+03,0.397988D+03, &
0.401411D+03,0.404862D+03,0.408326D+03,0.411829D+03,0.415352D+03/
      data (ES(IES),IES=856,950) / &
0.418906D+03,0.422490D+03,0.426095D+03,0.429740D+03,0.433398D+03, &
0.437097D+03,0.440827D+03,0.444570D+03,0.448354D+03,0.452160D+03, &
0.455999D+03,0.459870D+03,0.463765D+03,0.467702D+03,0.471652D+03, &
0.475646D+03,0.479674D+03,0.483715D+03,0.487811D+03,0.491911D+03, &
0.496065D+03,0.500244D+03,0.504448D+03,0.508698D+03,0.512961D+03, &
0.517282D+03,0.521617D+03,0.525989D+03,0.530397D+03,0.534831D+03, &
0.539313D+03,0.543821D+03,0.548355D+03,0.552938D+03,0.557549D+03, &
0.562197D+03,0.566884D+03,0.571598D+03,0.576351D+03,0.581131D+03, &
0.585963D+03,0.590835D+03,0.595722D+03,0.600663D+03,0.605631D+03, &
0.610641D+03,0.615151D+03,0.619625D+03,0.624140D+03,0.628671D+03, &
0.633243D+03,0.637845D+03,0.642465D+03,0.647126D+03,0.651806D+03, &
0.656527D+03,0.661279D+03,0.666049D+03,0.670861D+03,0.675692D+03, &
0.680566D+03,0.685471D+03,0.690396D+03,0.695363D+03,0.700350D+03, &
0.705381D+03,0.710444D+03,0.715527D+03,0.720654D+03,0.725801D+03, &
0.730994D+03,0.736219D+03,0.741465D+03,0.746756D+03,0.752068D+03, &
0.757426D+03,0.762819D+03,0.768231D+03,0.773692D+03,0.779172D+03, &
0.784701D+03,0.790265D+03,0.795849D+03,0.801483D+03,0.807137D+03, &
0.812842D+03,0.818582D+03,0.824343D+03,0.830153D+03,0.835987D+03, &
0.841871D+03,0.847791D+03,0.853733D+03,0.859727D+03,0.865743D+03/
      data (ES(IES),IES=951,1045) / &
0.871812D+03,0.877918D+03,0.884046D+03,0.890228D+03,0.896433D+03, &
0.902690D+03,0.908987D+03,0.915307D+03,0.921681D+03,0.928078D+03, &
0.934531D+03,0.941023D+03,0.947539D+03,0.954112D+03,0.960708D+03, &
0.967361D+03,0.974053D+03,0.980771D+03,0.987545D+03,0.994345D+03, &
0.100120D+04,0.100810D+04,0.101502D+04,0.102201D+04,0.102902D+04, &
0.103608D+04,0.104320D+04,0.105033D+04,0.105753D+04,0.106475D+04, &
0.107204D+04,0.107936D+04,0.108672D+04,0.109414D+04,0.110158D+04, &
0.110908D+04,0.111663D+04,0.112421D+04,0.113185D+04,0.113952D+04, &
0.114725D+04,0.115503D+04,0.116284D+04,0.117071D+04,0.117861D+04, &
0.118658D+04,0.119459D+04,0.120264D+04,0.121074D+04,0.121888D+04, &
0.122709D+04,0.123534D+04,0.124362D+04,0.125198D+04,0.126036D+04, &
0.126881D+04,0.127731D+04,0.128584D+04,0.129444D+04,0.130307D+04, &
0.131177D+04,0.132053D+04,0.132931D+04,0.133817D+04,0.134705D+04, &
0.135602D+04,0.136503D+04,0.137407D+04,0.138319D+04,0.139234D+04, &
0.140156D+04,0.141084D+04,0.142015D+04,0.142954D+04,0.143896D+04, &
0.144845D+04,0.145800D+04,0.146759D+04,0.147725D+04,0.148694D+04, &
0.149672D+04,0.150655D+04,0.151641D+04,0.152635D+04,0.153633D+04, &
0.154639D+04,0.155650D+04,0.156665D+04,0.157688D+04,0.158715D+04, &
0.159750D+04,0.160791D+04,0.161836D+04,0.162888D+04,0.163945D+04/
      data (ES(IES),IES=1046,1140) / &
0.165010D+04,0.166081D+04,0.167155D+04,0.168238D+04,0.169325D+04, &
0.170420D+04,0.171522D+04,0.172627D+04,0.173741D+04,0.174859D+04, &
0.175986D+04,0.177119D+04,0.178256D+04,0.179402D+04,0.180552D+04, &
0.181711D+04,0.182877D+04,0.184046D+04,0.185224D+04,0.186407D+04, &
0.187599D+04,0.188797D+04,0.190000D+04,0.191212D+04,0.192428D+04, &
0.193653D+04,0.194886D+04,0.196122D+04,0.197368D+04,0.198618D+04, &
0.199878D+04,0.201145D+04,0.202416D+04,0.203698D+04,0.204983D+04, &
0.206278D+04,0.207580D+04,0.208887D+04,0.210204D+04,0.211525D+04, &
0.212856D+04,0.214195D+04,0.215538D+04,0.216892D+04,0.218249D+04, &
0.219618D+04,0.220994D+04,0.222375D+04,0.223766D+04,0.225161D+04, &
0.226567D+04,0.227981D+04,0.229399D+04,0.230829D+04,0.232263D+04, &
0.233708D+04,0.235161D+04,0.236618D+04,0.238087D+04,0.239560D+04, &
0.241044D+04,0.242538D+04,0.244035D+04,0.245544D+04,0.247057D+04, &
0.248583D+04,0.250116D+04,0.251654D+04,0.253204D+04,0.254759D+04, &
0.256325D+04,0.257901D+04,0.259480D+04,0.261073D+04,0.262670D+04, &
0.264279D+04,0.265896D+04,0.267519D+04,0.269154D+04,0.270794D+04, &
0.272447D+04,0.274108D+04,0.275774D+04,0.277453D+04,0.279137D+04, &
0.280834D+04,0.282540D+04,0.284251D+04,0.285975D+04,0.287704D+04, &
0.289446D+04,0.291198D+04,0.292954D+04,0.294725D+04,0.296499D+04/
      data (ES(IES),IES=1141,1235) / &
0.298288D+04,0.300087D+04,0.301890D+04,0.303707D+04,0.305529D+04, &
0.307365D+04,0.309211D+04,0.311062D+04,0.312927D+04,0.314798D+04, &
0.316682D+04,0.318577D+04,0.320477D+04,0.322391D+04,0.324310D+04, &
0.326245D+04,0.328189D+04,0.330138D+04,0.332103D+04,0.334073D+04, &
0.336058D+04,0.338053D+04,0.340054D+04,0.342069D+04,0.344090D+04, &
0.346127D+04,0.348174D+04,0.350227D+04,0.352295D+04,0.354369D+04, &
0.356458D+04,0.358559D+04,0.360664D+04,0.362787D+04,0.364914D+04, &
0.367058D+04,0.369212D+04,0.371373D+04,0.373548D+04,0.375731D+04, &
0.377929D+04,0.380139D+04,0.382355D+04,0.384588D+04,0.386826D+04, &
0.389081D+04,0.391348D+04,0.393620D+04,0.395910D+04,0.398205D+04, &
0.400518D+04,0.402843D+04,0.405173D+04,0.407520D+04,0.409875D+04, &
0.412246D+04,0.414630D+04,0.417019D+04,0.419427D+04,0.421840D+04, &
0.424272D+04,0.426715D+04,0.429165D+04,0.431634D+04,0.434108D+04, &
0.436602D+04,0.439107D+04,0.441618D+04,0.444149D+04,0.446685D+04, &
0.449241D+04,0.451810D+04,0.454385D+04,0.456977D+04,0.459578D+04, &
0.462197D+04,0.464830D+04,0.467468D+04,0.470127D+04,0.472792D+04, &
0.475477D+04,0.478175D+04,0.480880D+04,0.483605D+04,0.486336D+04, &
0.489087D+04,0.491853D+04,0.494623D+04,0.497415D+04,0.500215D+04, &
0.503034D+04,0.505867D+04,0.508707D+04,0.511568D+04,0.514436D+04/
      data (ES(IES),IES=1236,1330) / &
0.517325D+04,0.520227D+04,0.523137D+04,0.526068D+04,0.529005D+04, &
0.531965D+04,0.534939D+04,0.537921D+04,0.540923D+04,0.543932D+04, &
0.546965D+04,0.550011D+04,0.553064D+04,0.556139D+04,0.559223D+04, &
0.562329D+04,0.565449D+04,0.568577D+04,0.571727D+04,0.574884D+04, &
0.578064D+04,0.581261D+04,0.584464D+04,0.587692D+04,0.590924D+04, &
0.594182D+04,0.597455D+04,0.600736D+04,0.604039D+04,0.607350D+04, &
0.610685D+04,0.614036D+04,0.617394D+04,0.620777D+04,0.624169D+04, &
0.627584D+04,0.631014D+04,0.634454D+04,0.637918D+04,0.641390D+04, &
0.644887D+04,0.648400D+04,0.651919D+04,0.655467D+04,0.659021D+04, &
0.662599D+04,0.666197D+04,0.669800D+04,0.673429D+04,0.677069D+04, &
0.680735D+04,0.684415D+04,0.688104D+04,0.691819D+04,0.695543D+04, &
0.699292D+04,0.703061D+04,0.706837D+04,0.710639D+04,0.714451D+04, &
0.718289D+04,0.722143D+04,0.726009D+04,0.729903D+04,0.733802D+04, &
0.737729D+04,0.741676D+04,0.745631D+04,0.749612D+04,0.753602D+04, &
0.757622D+04,0.761659D+04,0.765705D+04,0.769780D+04,0.773863D+04, &
0.777975D+04,0.782106D+04,0.786246D+04,0.790412D+04,0.794593D+04, &
0.798802D+04,0.803028D+04,0.807259D+04,0.811525D+04,0.815798D+04, &
0.820102D+04,0.824427D+04,0.828757D+04,0.833120D+04,0.837493D+04, &
0.841895D+04,0.846313D+04,0.850744D+04,0.855208D+04,0.859678D+04/
      data (ES(IES),IES=1331,1425) / &
0.864179D+04,0.868705D+04,0.873237D+04,0.877800D+04,0.882374D+04, &
0.886979D+04,0.891603D+04,0.896237D+04,0.900904D+04,0.905579D+04, &
0.910288D+04,0.915018D+04,0.919758D+04,0.924529D+04,0.929310D+04, &
0.934122D+04,0.938959D+04,0.943804D+04,0.948687D+04,0.953575D+04, &
0.958494D+04,0.963442D+04,0.968395D+04,0.973384D+04,0.978383D+04, &
0.983412D+04,0.988468D+04,0.993534D+04,0.998630D+04,0.100374D+05, &
0.100888D+05,0.101406D+05,0.101923D+05,0.102444D+05,0.102966D+05, &
0.103492D+05,0.104020D+05,0.104550D+05,0.105082D+05,0.105616D+05, &
0.106153D+05,0.106693D+05,0.107234D+05,0.107779D+05,0.108325D+05, &
0.108874D+05,0.109425D+05,0.109978D+05,0.110535D+05,0.111092D+05, &
0.111653D+05,0.112217D+05,0.112782D+05,0.113350D+05,0.113920D+05, &
0.114493D+05,0.115070D+05,0.115646D+05,0.116228D+05,0.116809D+05, &
0.117396D+05,0.117984D+05,0.118574D+05,0.119167D+05,0.119762D+05, &
0.120360D+05,0.120962D+05,0.121564D+05,0.122170D+05,0.122778D+05, &
0.123389D+05,0.124004D+05,0.124619D+05,0.125238D+05,0.125859D+05, &
0.126484D+05,0.127111D+05,0.127739D+05,0.128372D+05,0.129006D+05, &
0.129644D+05,0.130285D+05,0.130927D+05,0.131573D+05,0.132220D+05, &
0.132872D+05,0.133526D+05,0.134182D+05,0.134842D+05,0.135503D+05, &
0.136168D+05,0.136836D+05,0.137505D+05,0.138180D+05,0.138854D+05/
      data (ES(IES),IES=1426,1520) / &
0.139534D+05,0.140216D+05,0.140900D+05,0.141588D+05,0.142277D+05, &
0.142971D+05,0.143668D+05,0.144366D+05,0.145069D+05,0.145773D+05, &
0.146481D+05,0.147192D+05,0.147905D+05,0.148622D+05,0.149341D+05, &
0.150064D+05,0.150790D+05,0.151517D+05,0.152250D+05,0.152983D+05, &
0.153721D+05,0.154462D+05,0.155205D+05,0.155952D+05,0.156701D+05, &
0.157454D+05,0.158211D+05,0.158969D+05,0.159732D+05,0.160496D+05, &
0.161265D+05,0.162037D+05,0.162811D+05,0.163589D+05,0.164369D+05, &
0.165154D+05,0.165942D+05,0.166732D+05,0.167526D+05,0.168322D+05, &
0.169123D+05,0.169927D+05,0.170733D+05,0.171543D+05,0.172356D+05, &
0.173173D+05,0.173993D+05,0.174815D+05,0.175643D+05,0.176471D+05, &
0.177305D+05,0.178143D+05,0.178981D+05,0.179826D+05,0.180671D+05, &
0.181522D+05,0.182377D+05,0.183232D+05,0.184093D+05,0.184955D+05, &
0.185823D+05,0.186695D+05,0.187568D+05,0.188447D+05,0.189326D+05, &
0.190212D+05,0.191101D+05,0.191991D+05,0.192887D+05,0.193785D+05, &
0.194688D+05,0.195595D+05,0.196503D+05,0.197417D+05,0.198332D+05, &
0.199253D+05,0.200178D+05,0.201105D+05,0.202036D+05,0.202971D+05, &
0.203910D+05,0.204853D+05,0.205798D+05,0.206749D+05,0.207701D+05, &
0.208659D+05,0.209621D+05,0.210584D+05,0.211554D+05,0.212524D+05, &
0.213501D+05,0.214482D+05,0.215465D+05,0.216452D+05,0.217442D+05/
      data (ES(IES),IES=1521,1552) / &
0.218439D+05,0.219439D+05,0.220440D+05,0.221449D+05,0.222457D+05, &
0.223473D+05,0.224494D+05,0.225514D+05,0.226542D+05,0.227571D+05, &
0.228606D+05,0.229646D+05,0.230687D+05,0.231734D+05,0.232783D+05, &
0.233839D+05,0.234898D+05,0.235960D+05,0.237027D+05,0.238097D+05, &
0.239173D+05,0.240254D+05,0.241335D+05,0.242424D+05,0.243514D+05, &
0.244611D+05,0.245712D+05,0.246814D+05,0.247923D+05,0.249034D+05, &
0.250152D+05,0.250152D+05/

do I = 1, npnts

  ! Compute the factor that converts from sat vapour pressure in a
  ! pure water system to sat vapour pressure in air, FSUBW.  This formula
  ! is taken from equation A4.7 of Adrian Gill's book: Atmosphere-Ocean
  ! Dynamics.  Note that his formula works in terms of pressure in MB and
  ! temperature in Celsius, so conversion of units leads to the slightly
  ! different equation used here.

  FSUBW = one + 1.0E-8_kind_real * P(I) * &
               (4.5_kind_real+ 6.0E-4_kind_real * (T(I) - t0c) * (T(I) - t0c))

  ! use the lookup table to find saturated vapour pressure, and stored it in QS.

  TT = max (T_low, T(I))
  TT = min (T_high, TT)

  ATABLE = (TT - T_low + delta_T) / delta_T
  ITABLE = ATABLE
  ATABLE = ATABLE - ITABLE

  QS(I) = (one - ATABLE) * ES(ITABLE) + ATABLE * ES(ITABLE + 1)

  ! Multiply by FSUBW to convert to saturated vapour pressure in air (equation A4.6
  ! of Adrian Gill's book)

  QS(I) = QS(I) * FSUBW

  ! Now form the accurate expression for QS, which is a rearranged
  ! version of equation A4.3 of Gill's book.

  ! Note that at very low pressures we apply a fix, to prevent a
  ! singularity (Qsat tends to 1.0 kg/kg).

  QS(I) = (epsilon * QS(I)) / (max (P(I), QS(I)) - one_minus_epsilon * QS(I))

end do

end subroutine Ops_Qsat

!-------------------------------------------------------------------------------
!> Saturation Specific Humidity Scheme: Vapour to Liquid.
!!
!! \details Heritage: Ops_QsatWat.inc
!!
!! Returns a saturation mixing ratio given a temperature and pressure
!! using saturation vapour pressure calculated using the Goff-Gratch
!! Formulaie, adopted by the WMO as taken from Landolt-Bornstein, 1987
!! Numerical Data and Functional Relationships in Science and
!! Technology.  Group V/Vol 4B Meteorology.  Physical and Chemical
!! Properties of Air, P35
!!
!! Values in the lookup table are over water above and below 0 degrees C.
!!
!! Note: For vapour pressure oever water this formula is valid for
!! temperatures between 373K and 223K.  The values for saturated vapour
!! over water in the lookup table below are out of the lower end of
!! this range.  However it is standard WMO practice to use the formula
!! below its accepted range for use with the calculation of dew points
!! in the upper atmosphere
!!
!! Method: <br>
!!   uses lookup tables to find eSAT, calculates qSAT directly from that.
!!
!! \author Met Office
!!
!! \date 09/06/2020: Created
!!
subroutine Ops_QsatWat (QS, &
                         T, &
                         P, &
                       npnts)

implicit none

! subroutine arguments:
integer                     :: npnts     ! Points (=horizontal dimensions) being processed by qSAT scheme.
real(kind=kind_real)        :: T(npnts)  ! Temperature (K).
real(kind=kind_real)        :: P(npnts)  ! Pressure (Pa).
real(kind=kind_real)        :: QS(npnts) ! Saturation mixing ratio at temperature T and pressure P (KG/KG)

! Local declarations:
real(kind=kind_real), parameter :: one_minus_epsilon = one - epsilon
real(kind=kind_real), parameter :: T_low = 183.15_kind_real
real(kind=kind_real), parameter :: T_high = 338.15_kind_real
real(kind=kind_real), parameter :: delta_T = 0.1_kind_real
integer, parameter          :: N = ((T_high - T_low + (delta_T * half)) / delta_T) + one  ! gives N = 1551
integer                     :: ITABLE
real(kind=kind_real)        :: ATABLE
real(kind=kind_real)        :: FSUBW
real(kind=kind_real)        :: TT
integer                     :: I
integer                     :: IES
real(kind=kind_real)        :: ES(0:N + 1)   ! Table of saturation water vapour pressure (PA)
character(len=*), parameter :: RoutineName = "Ops_QsatWat"

! Note: 0 element is a repeat of 1st element to cater for special case
!       of low temperatures (.LE.T_low) for which the array index is
!       rounded down due to machine precision.

data (ES(IES),IES=    0, 95) / 0.186905D-01, &
0.186905D-01,0.190449D-01,0.194059D-01,0.197727D-01,0.201462D-01, &
0.205261D-01,0.209122D-01,0.213052D-01,0.217050D-01,0.221116D-01, &
0.225252D-01,0.229463D-01,0.233740D-01,0.238090D-01,0.242518D-01, &
0.247017D-01,0.251595D-01,0.256252D-01,0.260981D-01,0.265795D-01, &
0.270691D-01,0.275667D-01,0.280733D-01,0.285876D-01,0.291105D-01, &
0.296429D-01,0.301835D-01,0.307336D-01,0.312927D-01,0.318611D-01, &
0.324390D-01,0.330262D-01,0.336232D-01,0.342306D-01,0.348472D-01, &
0.354748D-01,0.361117D-01,0.367599D-01,0.374185D-01,0.380879D-01, &
0.387689D-01,0.394602D-01,0.401626D-01,0.408771D-01,0.416033D-01, &
0.423411D-01,0.430908D-01,0.438524D-01,0.446263D-01,0.454124D-01, &
0.462122D-01,0.470247D-01,0.478491D-01,0.486874D-01,0.495393D-01, &
0.504057D-01,0.512847D-01,0.521784D-01,0.530853D-01,0.540076D-01, &
0.549444D-01,0.558959D-01,0.568633D-01,0.578448D-01,0.588428D-01, &
0.598566D-01,0.608858D-01,0.619313D-01,0.629926D-01,0.640706D-01, &
0.651665D-01,0.662795D-01,0.674095D-01,0.685570D-01,0.697219D-01, &
0.709063D-01,0.721076D-01,0.733284D-01,0.745679D-01,0.758265D-01, &
0.771039D-01,0.784026D-01,0.797212D-01,0.810577D-01,0.824164D-01, &
0.837971D-01,0.851970D-01,0.866198D-01,0.880620D-01,0.895281D-01, &
0.910178D-01,0.925278D-01,0.940622D-01,0.956177D-01,0.971984D-01/
data (ES(IES),IES= 96,190) / &
0.988051D-01,0.100433D+00,0.102085D+00,0.103764D+00,0.105467D+00, &
0.107196D+00,0.108953D+00,0.110732D+00,0.112541D+00,0.114376D+00, &
0.116238D+00,0.118130D+00,0.120046D+00,0.121993D+00,0.123969D+00, &
0.125973D+00,0.128009D+00,0.130075D+00,0.132167D+00,0.134296D+00, &
0.136452D+00,0.138642D+00,0.140861D+00,0.143115D+00,0.145404D+00, &
0.147723D+00,0.150078D+00,0.152466D+00,0.154889D+00,0.157346D+00, &
0.159841D+00,0.162372D+00,0.164939D+00,0.167545D+00,0.170185D+00, &
0.172866D+00,0.175584D+00,0.178340D+00,0.181139D+00,0.183977D+00, &
0.186855D+00,0.189773D+00,0.192737D+00,0.195736D+00,0.198783D+00, &
0.201875D+00,0.205007D+00,0.208186D+00,0.211409D+00,0.214676D+00, &
0.217993D+00,0.221355D+00,0.224764D+00,0.228220D+00,0.231728D+00, &
0.235284D+00,0.238888D+00,0.242542D+00,0.246251D+00,0.250010D+00, &
0.253821D+00,0.257688D+00,0.261602D+00,0.265575D+00,0.269607D+00, &
0.273689D+00,0.277830D+00,0.282027D+00,0.286287D+00,0.290598D+00, &
0.294972D+00,0.299405D+00,0.303904D+00,0.308462D+00,0.313082D+00, &
0.317763D+00,0.322512D+00,0.327324D+00,0.332201D+00,0.337141D+00, &
0.342154D+00,0.347234D+00,0.352387D+00,0.357601D+00,0.362889D+00, &
0.368257D+00,0.373685D+00,0.379194D+00,0.384773D+00,0.390433D+00, &
0.396159D+00,0.401968D+00,0.407861D+00,0.413820D+00,0.419866D+00/
data (ES(IES),IES=191,285) / &
0.425999D+00,0.432203D+00,0.438494D+00,0.444867D+00,0.451332D+00, &
0.457879D+00,0.464510D+00,0.471226D+00,0.478037D+00,0.484935D+00, &
0.491920D+00,0.499005D+00,0.506181D+00,0.513447D+00,0.520816D+00, &
0.528279D+00,0.535835D+00,0.543497D+00,0.551256D+00,0.559113D+00, &
0.567081D+00,0.575147D+00,0.583315D+00,0.591585D+00,0.599970D+00, &
0.608472D+00,0.617069D+00,0.625785D+00,0.634609D+00,0.643556D+00, &
0.652611D+00,0.661782D+00,0.671077D+00,0.680487D+00,0.690015D+00, &
0.699656D+00,0.709433D+00,0.719344D+00,0.729363D+00,0.739518D+00, &
0.749795D+00,0.760217D+00,0.770763D+00,0.781454D+00,0.792258D+00, &
0.803208D+00,0.814309D+00,0.825528D+00,0.836914D+00,0.848422D+00, &
0.860086D+00,0.871891D+00,0.883837D+00,0.895944D+00,0.908214D+00, &
0.920611D+00,0.933175D+00,0.945890D+00,0.958776D+00,0.971812D+00, &
0.985027D+00,0.998379D+00,0.101193D+01,0.102561D+01,0.103949D+01, &
0.105352D+01,0.106774D+01,0.108213D+01,0.109669D+01,0.111144D+01, &
0.112636D+01,0.114148D+01,0.115676D+01,0.117226D+01,0.118791D+01, &
0.120377D+01,0.121984D+01,0.123608D+01,0.125252D+01,0.126919D+01, &
0.128604D+01,0.130309D+01,0.132036D+01,0.133782D+01,0.135549D+01, &
0.137339D+01,0.139150D+01,0.140984D+01,0.142839D+01,0.144715D+01, &
0.146616D+01,0.148538D+01,0.150482D+01,0.152450D+01,0.154445D+01/
data (ES(IES),IES=286,380) / &
0.156459D+01,0.158502D+01,0.160564D+01,0.162654D+01,0.164766D+01, &
0.166906D+01,0.169070D+01,0.171257D+01,0.173473D+01,0.175718D+01, &
0.177984D+01,0.180282D+01,0.182602D+01,0.184951D+01,0.187327D+01, &
0.189733D+01,0.192165D+01,0.194629D+01,0.197118D+01,0.199636D+01, &
0.202185D+01,0.204762D+01,0.207372D+01,0.210010D+01,0.212678D+01, &
0.215379D+01,0.218109D+01,0.220873D+01,0.223668D+01,0.226497D+01, &
0.229357D+01,0.232249D+01,0.235176D+01,0.238134D+01,0.241129D+01, &
0.244157D+01,0.247217D+01,0.250316D+01,0.253447D+01,0.256617D+01, &
0.259821D+01,0.263064D+01,0.266341D+01,0.269661D+01,0.273009D+01, &
0.276403D+01,0.279834D+01,0.283302D+01,0.286811D+01,0.290358D+01, &
0.293943D+01,0.297571D+01,0.301236D+01,0.304946D+01,0.308702D+01, &
0.312491D+01,0.316326D+01,0.320208D+01,0.324130D+01,0.328092D+01, &
0.332102D+01,0.336162D+01,0.340264D+01,0.344407D+01,0.348601D+01, &
0.352838D+01,0.357118D+01,0.361449D+01,0.365834D+01,0.370264D+01, &
0.374737D+01,0.379265D+01,0.383839D+01,0.388469D+01,0.393144D+01, &
0.397876D+01,0.402656D+01,0.407492D+01,0.412378D+01,0.417313D+01, &
0.422306D+01,0.427359D+01,0.432454D+01,0.437617D+01,0.442834D+01, &
0.448102D+01,0.453433D+01,0.458816D+01,0.464253D+01,0.469764D+01, &
0.475321D+01,0.480942D+01,0.486629D+01,0.492372D+01,0.498173D+01/
data (ES(IES),IES=381,475) / &
0.504041D+01,0.509967D+01,0.515962D+01,0.522029D+01,0.528142D+01, &
0.534337D+01,0.540595D+01,0.546912D+01,0.553292D+01,0.559757D+01, &
0.566273D+01,0.572864D+01,0.579532D+01,0.586266D+01,0.593075D+01, &
0.599952D+01,0.606895D+01,0.613918D+01,0.621021D+01,0.628191D+01, &
0.635433D+01,0.642755D+01,0.650162D+01,0.657639D+01,0.665188D+01, &
0.672823D+01,0.680532D+01,0.688329D+01,0.696198D+01,0.704157D+01, &
0.712206D+01,0.720319D+01,0.728534D+01,0.736829D+01,0.745204D+01, &
0.753671D+01,0.762218D+01,0.770860D+01,0.779588D+01,0.788408D+01, &
0.797314D+01,0.806318D+01,0.815408D+01,0.824599D+01,0.833874D+01, &
0.843254D+01,0.852721D+01,0.862293D+01,0.871954D+01,0.881724D+01, &
0.891579D+01,0.901547D+01,0.911624D+01,0.921778D+01,0.932061D+01, &
0.942438D+01,0.952910D+01,0.963497D+01,0.974181D+01,0.984982D+01, &
0.995887D+01,0.100690D+02,0.101804D+02,0.102926D+02,0.104063D+02, &
0.105210D+02,0.106367D+02,0.107536D+02,0.108719D+02,0.109912D+02, &
0.111116D+02,0.112333D+02,0.113563D+02,0.114804D+02,0.116056D+02, &
0.117325D+02,0.118602D+02,0.119892D+02,0.121197D+02,0.122513D+02, &
0.123844D+02,0.125186D+02,0.126543D+02,0.127912D+02,0.129295D+02, &
0.130691D+02,0.132101D+02,0.133527D+02,0.134965D+02,0.136415D+02, &
0.137882D+02,0.139361D+02,0.140855D+02,0.142366D+02,0.143889D+02/
data (ES(IES),IES=476,570) / &
0.145429D+02,0.146982D+02,0.148552D+02,0.150135D+02,0.151735D+02, &
0.153349D+02,0.154979D+02,0.156624D+02,0.158286D+02,0.159965D+02, &
0.161659D+02,0.163367D+02,0.165094D+02,0.166838D+02,0.168597D+02, &
0.170375D+02,0.172168D+02,0.173979D+02,0.175806D+02,0.177651D+02, &
0.179513D+02,0.181394D+02,0.183293D+02,0.185210D+02,0.187146D+02, &
0.189098D+02,0.191066D+02,0.193059D+02,0.195065D+02,0.197095D+02, &
0.199142D+02,0.201206D+02,0.203291D+02,0.205397D+02,0.207522D+02, &
0.209664D+02,0.211831D+02,0.214013D+02,0.216221D+02,0.218448D+02, &
0.220692D+02,0.222959D+02,0.225250D+02,0.227559D+02,0.229887D+02, &
0.232239D+02,0.234614D+02,0.237014D+02,0.239428D+02,0.241872D+02, &
0.244335D+02,0.246824D+02,0.249332D+02,0.251860D+02,0.254419D+02, &
0.256993D+02,0.259600D+02,0.262225D+02,0.264873D+02,0.267552D+02, &
0.270248D+02,0.272970D+02,0.275719D+02,0.278497D+02,0.281295D+02, &
0.284117D+02,0.286965D+02,0.289843D+02,0.292743D+02,0.295671D+02, &
0.298624D+02,0.301605D+02,0.304616D+02,0.307650D+02,0.310708D+02, &
0.313803D+02,0.316915D+02,0.320064D+02,0.323238D+02,0.326437D+02, &
0.329666D+02,0.332928D+02,0.336215D+02,0.339534D+02,0.342885D+02, &
0.346263D+02,0.349666D+02,0.353109D+02,0.356572D+02,0.360076D+02, &
0.363606D+02,0.367164D+02,0.370757D+02,0.374383D+02,0.378038D+02/
data (ES(IES),IES=571,665) / &
0.381727D+02,0.385453D+02,0.389206D+02,0.392989D+02,0.396807D+02, &
0.400663D+02,0.404555D+02,0.408478D+02,0.412428D+02,0.416417D+02, &
0.420445D+02,0.424502D+02,0.428600D+02,0.432733D+02,0.436900D+02, &
0.441106D+02,0.445343D+02,0.449620D+02,0.453930D+02,0.458280D+02, &
0.462672D+02,0.467096D+02,0.471561D+02,0.476070D+02,0.480610D+02, &
0.485186D+02,0.489813D+02,0.494474D+02,0.499170D+02,0.503909D+02, &
0.508693D+02,0.513511D+02,0.518376D+02,0.523277D+02,0.528232D+02, &
0.533213D+02,0.538240D+02,0.543315D+02,0.548437D+02,0.553596D+02, &
0.558802D+02,0.564046D+02,0.569340D+02,0.574672D+02,0.580061D+02, &
0.585481D+02,0.590963D+02,0.596482D+02,0.602041D+02,0.607649D+02, &
0.613311D+02,0.619025D+02,0.624779D+02,0.630574D+02,0.636422D+02, &
0.642324D+02,0.648280D+02,0.654278D+02,0.660332D+02,0.666426D+02, &
0.672577D+02,0.678771D+02,0.685034D+02,0.691328D+02,0.697694D+02, &
0.704103D+02,0.710556D+02,0.717081D+02,0.723639D+02,0.730269D+02, &
0.736945D+02,0.743681D+02,0.750463D+02,0.757309D+02,0.764214D+02, &
0.771167D+02,0.778182D+02,0.785246D+02,0.792373D+02,0.799564D+02, &
0.806804D+02,0.814109D+02,0.821479D+02,0.828898D+02,0.836384D+02, &
0.843922D+02,0.851525D+02,0.859198D+02,0.866920D+02,0.874712D+02, &
0.882574D+02,0.890486D+02,0.898470D+02,0.906525D+02,0.914634D+02/
data (ES(IES),IES=666,760) / &
0.922814D+02,0.931048D+02,0.939356D+02,0.947736D+02,0.956171D+02, &
0.964681D+02,0.973246D+02,0.981907D+02,0.990605D+02,0.999399D+02, &
0.100825D+03,0.101718D+03,0.102617D+03,0.103523D+03,0.104438D+03, &
0.105358D+03,0.106287D+03,0.107221D+03,0.108166D+03,0.109115D+03, &
0.110074D+03,0.111039D+03,0.112012D+03,0.112992D+03,0.113981D+03, &
0.114978D+03,0.115981D+03,0.116993D+03,0.118013D+03,0.119041D+03, &
0.120077D+03,0.121122D+03,0.122173D+03,0.123234D+03,0.124301D+03, &
0.125377D+03,0.126463D+03,0.127556D+03,0.128657D+03,0.129769D+03, &
0.130889D+03,0.132017D+03,0.133152D+03,0.134299D+03,0.135453D+03, &
0.136614D+03,0.137786D+03,0.138967D+03,0.140158D+03,0.141356D+03, &
0.142565D+03,0.143781D+03,0.145010D+03,0.146247D+03,0.147491D+03, &
0.148746D+03,0.150011D+03,0.151284D+03,0.152571D+03,0.153862D+03, &
0.155168D+03,0.156481D+03,0.157805D+03,0.159137D+03,0.160478D+03, &
0.161832D+03,0.163198D+03,0.164569D+03,0.165958D+03,0.167348D+03, &
0.168757D+03,0.170174D+03,0.171599D+03,0.173037D+03,0.174483D+03, &
0.175944D+03,0.177414D+03,0.178892D+03,0.180387D+03,0.181886D+03, &
0.183402D+03,0.184930D+03,0.186463D+03,0.188012D+03,0.189571D+03, &
0.191146D+03,0.192730D+03,0.194320D+03,0.195930D+03,0.197546D+03, &
0.199175D+03,0.200821D+03,0.202473D+03,0.204142D+03,0.205817D+03/
data (ES(IES),IES=761,855) / &
0.207510D+03,0.209216D+03,0.210928D+03,0.212658D+03,0.214398D+03, &
0.216152D+03,0.217920D+03,0.219698D+03,0.221495D+03,0.223297D+03, &
0.225119D+03,0.226951D+03,0.228793D+03,0.230654D+03,0.232522D+03, &
0.234413D+03,0.236311D+03,0.238223D+03,0.240151D+03,0.242090D+03, &
0.244049D+03,0.246019D+03,0.248000D+03,0.249996D+03,0.252009D+03, &
0.254037D+03,0.256077D+03,0.258128D+03,0.260200D+03,0.262284D+03, &
0.264384D+03,0.266500D+03,0.268629D+03,0.270779D+03,0.272936D+03, &
0.275110D+03,0.277306D+03,0.279509D+03,0.281734D+03,0.283966D+03, &
0.286227D+03,0.288494D+03,0.290780D+03,0.293083D+03,0.295398D+03, &
0.297737D+03,0.300089D+03,0.302453D+03,0.304841D+03,0.307237D+03, &
0.309656D+03,0.312095D+03,0.314541D+03,0.317012D+03,0.319496D+03, &
0.322005D+03,0.324527D+03,0.327063D+03,0.329618D+03,0.332193D+03, &
0.334788D+03,0.337396D+03,0.340025D+03,0.342673D+03,0.345329D+03, &
0.348019D+03,0.350722D+03,0.353440D+03,0.356178D+03,0.358938D+03, &
0.361718D+03,0.364513D+03,0.367322D+03,0.370160D+03,0.373012D+03, &
0.375885D+03,0.378788D+03,0.381691D+03,0.384631D+03,0.387579D+03, &
0.390556D+03,0.393556D+03,0.396563D+03,0.399601D+03,0.402646D+03, &
0.405730D+03,0.408829D+03,0.411944D+03,0.415083D+03,0.418236D+03, &
0.421422D+03,0.424632D+03,0.427849D+03,0.431099D+03,0.434365D+03/
data (ES(IES),IES=856,950) / &
0.437655D+03,0.440970D+03,0.444301D+03,0.447666D+03,0.451038D+03, &
0.454445D+03,0.457876D+03,0.461316D+03,0.464790D+03,0.468281D+03, &
0.471798D+03,0.475342D+03,0.478902D+03,0.482497D+03,0.486101D+03, &
0.489741D+03,0.493408D+03,0.497083D+03,0.500804D+03,0.504524D+03, &
0.508290D+03,0.512074D+03,0.515877D+03,0.519717D+03,0.523566D+03, &
0.527462D+03,0.531367D+03,0.535301D+03,0.539264D+03,0.543245D+03, &
0.547265D+03,0.551305D+03,0.555363D+03,0.559462D+03,0.563579D+03, &
0.567727D+03,0.571905D+03,0.576102D+03,0.580329D+03,0.584576D+03, &
0.588865D+03,0.593185D+03,0.597514D+03,0.601885D+03,0.606276D+03, &
0.610699D+03,0.615151D+03,0.619625D+03,0.624140D+03,0.628671D+03, &
0.633243D+03,0.637845D+03,0.642465D+03,0.647126D+03,0.651806D+03, &
0.656527D+03,0.661279D+03,0.666049D+03,0.670861D+03,0.675692D+03, &
0.680566D+03,0.685471D+03,0.690396D+03,0.695363D+03,0.700350D+03, &
0.705381D+03,0.710444D+03,0.715527D+03,0.720654D+03,0.725801D+03, &
0.730994D+03,0.736219D+03,0.741465D+03,0.746756D+03,0.752068D+03, &
0.757426D+03,0.762819D+03,0.768231D+03,0.773692D+03,0.779172D+03, &
0.784701D+03,0.790265D+03,0.795849D+03,0.801483D+03,0.807137D+03, &
0.812842D+03,0.818582D+03,0.824343D+03,0.830153D+03,0.835987D+03, &
0.841871D+03,0.847791D+03,0.853733D+03,0.859727D+03,0.865743D+03/
data (ES(IES),IES=951,1045) / &
0.871812D+03,0.877918D+03,0.884046D+03,0.890228D+03,0.896433D+03, &
0.902690D+03,0.908987D+03,0.915307D+03,0.921681D+03,0.928078D+03, &
0.934531D+03,0.941023D+03,0.947539D+03,0.954112D+03,0.960708D+03, &
0.967361D+03,0.974053D+03,0.980771D+03,0.987545D+03,0.994345D+03, &
0.100120D+04,0.100810D+04,0.101502D+04,0.102201D+04,0.102902D+04, &
0.103608D+04,0.104320D+04,0.105033D+04,0.105753D+04,0.106475D+04, &
0.107204D+04,0.107936D+04,0.108672D+04,0.109414D+04,0.110158D+04, &
0.110908D+04,0.111663D+04,0.112421D+04,0.113185D+04,0.113952D+04, &
0.114725D+04,0.115503D+04,0.116284D+04,0.117071D+04,0.117861D+04, &
0.118658D+04,0.119459D+04,0.120264D+04,0.121074D+04,0.121888D+04, &
0.122709D+04,0.123534D+04,0.124362D+04,0.125198D+04,0.126036D+04, &
0.126881D+04,0.127731D+04,0.128584D+04,0.129444D+04,0.130307D+04, &
0.131177D+04,0.132053D+04,0.132931D+04,0.133817D+04,0.134705D+04, &
0.135602D+04,0.136503D+04,0.137407D+04,0.138319D+04,0.139234D+04, &
0.140156D+04,0.141084D+04,0.142015D+04,0.142954D+04,0.143896D+04, &
0.144845D+04,0.145800D+04,0.146759D+04,0.147725D+04,0.148694D+04, &
0.149672D+04,0.150655D+04,0.151641D+04,0.152635D+04,0.153633D+04, &
0.154639D+04,0.155650D+04,0.156665D+04,0.157688D+04,0.158715D+04, &
0.159750D+04,0.160791D+04,0.161836D+04,0.162888D+04,0.163945D+04/
data (ES(IES),IES=1046,1140) / &
0.165010D+04,0.166081D+04,0.167155D+04,0.168238D+04,0.169325D+04, &
0.170420D+04,0.171522D+04,0.172627D+04,0.173741D+04,0.174859D+04, &
0.175986D+04,0.177119D+04,0.178256D+04,0.179402D+04,0.180552D+04, &
0.181711D+04,0.182877D+04,0.184046D+04,0.185224D+04,0.186407D+04, &
0.187599D+04,0.188797D+04,0.190000D+04,0.191212D+04,0.192428D+04, &
0.193653D+04,0.194886D+04,0.196122D+04,0.197368D+04,0.198618D+04, &
0.199878D+04,0.201145D+04,0.202416D+04,0.203698D+04,0.204983D+04, &
0.206278D+04,0.207580D+04,0.208887D+04,0.210204D+04,0.211525D+04, &
0.212856D+04,0.214195D+04,0.215538D+04,0.216892D+04,0.218249D+04, &
0.219618D+04,0.220994D+04,0.222375D+04,0.223766D+04,0.225161D+04, &
0.226567D+04,0.227981D+04,0.229399D+04,0.230829D+04,0.232263D+04, &
0.233708D+04,0.235161D+04,0.236618D+04,0.238087D+04,0.239560D+04, &
0.241044D+04,0.242538D+04,0.244035D+04,0.245544D+04,0.247057D+04, &
0.248583D+04,0.250116D+04,0.251654D+04,0.253204D+04,0.254759D+04, &
0.256325D+04,0.257901D+04,0.259480D+04,0.261073D+04,0.262670D+04, &
0.264279D+04,0.265896D+04,0.267519D+04,0.269154D+04,0.270794D+04, &
0.272447D+04,0.274108D+04,0.275774D+04,0.277453D+04,0.279137D+04, &
0.280834D+04,0.282540D+04,0.284251D+04,0.285975D+04,0.287704D+04, &
0.289446D+04,0.291198D+04,0.292954D+04,0.294725D+04,0.296499D+04/
data (ES(IES),IES=1141,1235) / &
0.298288D+04,0.300087D+04,0.301890D+04,0.303707D+04,0.305529D+04, &
0.307365D+04,0.309211D+04,0.311062D+04,0.312927D+04,0.314798D+04, &
0.316682D+04,0.318577D+04,0.320477D+04,0.322391D+04,0.324310D+04, &
0.326245D+04,0.328189D+04,0.330138D+04,0.332103D+04,0.334073D+04, &
0.336058D+04,0.338053D+04,0.340054D+04,0.342069D+04,0.344090D+04, &
0.346127D+04,0.348174D+04,0.350227D+04,0.352295D+04,0.354369D+04, &
0.356458D+04,0.358559D+04,0.360664D+04,0.362787D+04,0.364914D+04, &
0.367058D+04,0.369212D+04,0.371373D+04,0.373548D+04,0.375731D+04, &
0.377929D+04,0.380139D+04,0.382355D+04,0.384588D+04,0.386826D+04, &
0.389081D+04,0.391348D+04,0.393620D+04,0.395910D+04,0.398205D+04, &
0.400518D+04,0.402843D+04,0.405173D+04,0.407520D+04,0.409875D+04, &
0.412246D+04,0.414630D+04,0.417019D+04,0.419427D+04,0.421840D+04, &
0.424272D+04,0.426715D+04,0.429165D+04,0.431634D+04,0.434108D+04, &
0.436602D+04,0.439107D+04,0.441618D+04,0.444149D+04,0.446685D+04, &
0.449241D+04,0.451810D+04,0.454385D+04,0.456977D+04,0.459578D+04, &
0.462197D+04,0.464830D+04,0.467468D+04,0.470127D+04,0.472792D+04, &
0.475477D+04,0.478175D+04,0.480880D+04,0.483605D+04,0.486336D+04, &
0.489087D+04,0.491853D+04,0.494623D+04,0.497415D+04,0.500215D+04, &
0.503034D+04,0.505867D+04,0.508707D+04,0.511568D+04,0.514436D+04/
data (ES(IES),IES=1236,1330) / &
0.517325D+04,0.520227D+04,0.523137D+04,0.526068D+04,0.529005D+04, &
0.531965D+04,0.534939D+04,0.537921D+04,0.540923D+04,0.543932D+04, &
0.546965D+04,0.550011D+04,0.553064D+04,0.556139D+04,0.559223D+04, &
0.562329D+04,0.565449D+04,0.568577D+04,0.571727D+04,0.574884D+04, &
0.578064D+04,0.581261D+04,0.584464D+04,0.587692D+04,0.590924D+04, &
0.594182D+04,0.597455D+04,0.600736D+04,0.604039D+04,0.607350D+04, &
0.610685D+04,0.614036D+04,0.617394D+04,0.620777D+04,0.624169D+04, &
0.627584D+04,0.631014D+04,0.634454D+04,0.637918D+04,0.641390D+04, &
0.644887D+04,0.648400D+04,0.651919D+04,0.655467D+04,0.659021D+04, &
0.662599D+04,0.666197D+04,0.669800D+04,0.673429D+04,0.677069D+04, &
0.680735D+04,0.684415D+04,0.688104D+04,0.691819D+04,0.695543D+04, &
0.699292D+04,0.703061D+04,0.706837D+04,0.710639D+04,0.714451D+04, &
0.718289D+04,0.722143D+04,0.726009D+04,0.729903D+04,0.733802D+04, &
0.737729D+04,0.741676D+04,0.745631D+04,0.749612D+04,0.753602D+04, &
0.757622D+04,0.761659D+04,0.765705D+04,0.769780D+04,0.773863D+04, &
0.777975D+04,0.782106D+04,0.786246D+04,0.790412D+04,0.794593D+04, &
0.798802D+04,0.803028D+04,0.807259D+04,0.811525D+04,0.815798D+04, &
0.820102D+04,0.824427D+04,0.828757D+04,0.833120D+04,0.837493D+04, &
0.841895D+04,0.846313D+04,0.850744D+04,0.855208D+04,0.859678D+04/
data (ES(IES),IES=1331,1425) / &
0.864179D+04,0.868705D+04,0.873237D+04,0.877800D+04,0.882374D+04, &
0.886979D+04,0.891603D+04,0.896237D+04,0.900904D+04,0.905579D+04, &
0.910288D+04,0.915018D+04,0.919758D+04,0.924529D+04,0.929310D+04, &
0.934122D+04,0.938959D+04,0.943804D+04,0.948687D+04,0.953575D+04, &
0.958494D+04,0.963442D+04,0.968395D+04,0.973384D+04,0.978383D+04, &
0.983412D+04,0.988468D+04,0.993534D+04,0.998630D+04,0.100374D+05, &
0.100888D+05,0.101406D+05,0.101923D+05,0.102444D+05,0.102966D+05, &
0.103492D+05,0.104020D+05,0.104550D+05,0.105082D+05,0.105616D+05, &
0.106153D+05,0.106693D+05,0.107234D+05,0.107779D+05,0.108325D+05, &
0.108874D+05,0.109425D+05,0.109978D+05,0.110535D+05,0.111092D+05, &
0.111653D+05,0.112217D+05,0.112782D+05,0.113350D+05,0.113920D+05, &
0.114493D+05,0.115070D+05,0.115646D+05,0.116228D+05,0.116809D+05, &
0.117396D+05,0.117984D+05,0.118574D+05,0.119167D+05,0.119762D+05, &
0.120360D+05,0.120962D+05,0.121564D+05,0.122170D+05,0.122778D+05, &
0.123389D+05,0.124004D+05,0.124619D+05,0.125238D+05,0.125859D+05, &
0.126484D+05,0.127111D+05,0.127739D+05,0.128372D+05,0.129006D+05, &
0.129644D+05,0.130285D+05,0.130927D+05,0.131573D+05,0.132220D+05, &
0.132872D+05,0.133526D+05,0.134182D+05,0.134842D+05,0.135503D+05, &
0.136168D+05,0.136836D+05,0.137505D+05,0.138180D+05,0.138854D+05/
data (ES(IES),IES=1426,1520) / &
0.139534D+05,0.140216D+05,0.140900D+05,0.141588D+05,0.142277D+05, &
0.142971D+05,0.143668D+05,0.144366D+05,0.145069D+05,0.145773D+05, &
0.146481D+05,0.147192D+05,0.147905D+05,0.148622D+05,0.149341D+05, &
0.150064D+05,0.150790D+05,0.151517D+05,0.152250D+05,0.152983D+05, &
0.153721D+05,0.154462D+05,0.155205D+05,0.155952D+05,0.156701D+05, &
0.157454D+05,0.158211D+05,0.158969D+05,0.159732D+05,0.160496D+05, &
0.161265D+05,0.162037D+05,0.162811D+05,0.163589D+05,0.164369D+05, &
0.165154D+05,0.165942D+05,0.166732D+05,0.167526D+05,0.168322D+05, &
0.169123D+05,0.169927D+05,0.170733D+05,0.171543D+05,0.172356D+05, &
0.173173D+05,0.173993D+05,0.174815D+05,0.175643D+05,0.176471D+05, &
0.177305D+05,0.178143D+05,0.178981D+05,0.179826D+05,0.180671D+05, &
0.181522D+05,0.182377D+05,0.183232D+05,0.184093D+05,0.184955D+05, &
0.185823D+05,0.186695D+05,0.187568D+05,0.188447D+05,0.189326D+05, &
0.190212D+05,0.191101D+05,0.191991D+05,0.192887D+05,0.193785D+05, &
0.194688D+05,0.195595D+05,0.196503D+05,0.197417D+05,0.198332D+05, &
0.199253D+05,0.200178D+05,0.201105D+05,0.202036D+05,0.202971D+05, &
0.203910D+05,0.204853D+05,0.205798D+05,0.206749D+05,0.207701D+05, &
0.208659D+05,0.209621D+05,0.210584D+05,0.211554D+05,0.212524D+05, &
0.213501D+05,0.214482D+05,0.215465D+05,0.216452D+05,0.217442D+05/
data (ES(IES),IES=1521,1552) / &
0.218439D+05,0.219439D+05,0.220440D+05,0.221449D+05,0.222457D+05, &
0.223473D+05,0.224494D+05,0.225514D+05,0.226542D+05,0.227571D+05, &
0.228606D+05,0.229646D+05,0.230687D+05,0.231734D+05,0.232783D+05, &
0.233839D+05,0.234898D+05,0.235960D+05,0.237027D+05,0.238097D+05, &
0.239173D+05,0.240254D+05,0.241335D+05,0.242424D+05,0.243514D+05, &
0.244611D+05,0.245712D+05,0.246814D+05,0.247923D+05,0.249034D+05, &
0.250152D+05,0.250152D+05/

do I = 1, npnts

  ! Compute the factor that converts from sat vapour pressure in a
  ! pure water system to sat vapour pressure in air, FSUBW.  This formula
  ! is taken from equation A4.7 of Adrian Gill's Book: Atmosphere-Ocean
  ! Dynamics.  Note that his formula works in terms of pressure in MB and
  ! temperature in Celsius, so conversion of units leads to the slightly
  ! different equation used here.

  FSUBW = one + 1.0E-8_kind_real * P(I) &
          * (4.5_kind_real + 6.0E-4_kind_real * (T(I) - t0c) * (T(I) - t0c))

  ! use the lookup table to find saturated vapour pressure, and store it in QS

  TT = max (T_low, T(I))
  TT = min (T_high, TT)

  ATABLE = (TT - T_low + delta_T) / delta_T
  ITABLE = ATABLE
  ATABLE = ATABLE - ITABLE

  QS(I) = (one - ATABLE) * ES(ITABLE) + ATABLE * ES(ITABLE + 1)

  ! Multiple by FSUBW to convert to saturated vapour pressure in air
  ! (equation A4.6 of Adrian Gill's book)

  QS(I) = QS(I) * FSUBW

  ! Now form the accurate expression for QS, which is a rearranged version of
  ! equation A4.3 of Gill's book.
  !
  ! Note that at very low pressures we apply a fix, to prevent a singularity
  ! (Qsat tends to 1.0 kg/kg)

  QS(I) = (epsilon * QS(I)) / (max (P(I), QS(I)) - one_minus_epsilon * QS(I))

end do

end subroutine Ops_QsatWat

!-------------------------------------------------------------------------------
!> Split the humidity into water vapour, liquid water and ice.
!!
!! \details Heritage: Ops_SatRad_Qsplit.f90
!!
!! if output_type=1 : Split total water content (qtotal) into <br>
!!   water vapor content (q) and <br>
!!   cloud liquid water content (ql) and <br>
!!   cloud ice water content (qi) <br>
!!
!! if output_type ne 1 : Compute derivatives: (q) =dq/dqtotal <br>
!!                                            (ql)=dql/dqtotal <br>
!!                                            (qi)=dqi/dqtotal <br>
!!
!! \warning The derivatives are not valid if LtemperatureVar=.true. since
!!     qsaturated depends on temperature.
!!
!! The partitioning of the excess moisture between ice and clw uses a temperature
!! based parametrization based on aircraft data Ref: Jones DC Reading phdthesis
!! p126
!!
!! \author Met Office
!!
!! \date 09/06/2020: Created
!!
subroutine Ops_SatRad_Qsplit ( &
  output_type, &
  p,           &
  t,           &
  qtotal,      &
  q,           &
  ql,          &
  qi,          &
  UseQtSplitRain)

  implicit none

  ! Subroutine arguments:
  integer, intent(in)                    :: output_type
  real(kind_real), intent(in)            :: p(:)
  real(kind_real), intent(in)            :: t(:)
  real(kind_real), intent(in)            :: qtotal(:)
  real(kind_real), intent(out)           :: q(size(qtotal))      ! humidity component q
  real(kind_real), intent(out)           :: ql(size(qtotal))     ! liquid component ql
  real(kind_real), intent(out)           :: qi(size(qtotal))     ! ice component qi
  logical,         intent(in)            :: UseQtSplitRain

  ! Local declarations:
  integer                                :: i
  real(kind_real), parameter             :: lower_rh = 0.95_kind_real
  real(kind_real), parameter             :: upper_rh = 1.05_kind_real
  real(kind_real), parameter             :: Split_Factor = half
  real(kind_real), parameter             :: MinTempQl = 233.15_kind_real ! temperature (K) below which all cloud is ice
  real(kind_real)                        :: qsaturated(size(qtotal))
  real(kind_real)                        :: RH_qtotal(size(qtotal))
  real(kind_real)                        :: qnv(size(qtotal))        ! non vapour component
  real(kind_real)                        :: qc(size(qtotal))         ! cloud component
  real(kind_real)                        :: V1(size(qtotal))
  real(kind_real)                        :: V2(size(qtotal))
  real(kind_real)                        :: V1zero
  real(kind_real)                        :: V2zero
  real(kind_real)                        :: W(size(qtotal))
  real(kind_real)                        :: Y1,Y2,Y3,Y4
  real(kind_real)                        :: IntConst
  real(kind_real)                        :: Aconst
  real(kind_real)                        :: Bconst
  real(kind_real)                        :: Cconst
  real(kind_real)                        :: Dconst
  real(kind_real)                        :: Denom
  real(kind_real)                        :: SmallValue
  real(kind_real)                        :: LF(size(qtotal))          ! fraction of ql to ql+qi
  character(len=*), parameter :: RoutineName = "Ops_SatRad_Qsplit"

  real(kind_real)    :: QsplitRainParamA       !Parameters used to define proportion of
  real(kind_real)    :: QsplitRainParamB       !qt that is partitioned into a rain compnenent
  real(kind_real)    :: QsplitRainParamC       !

  !Qsplit_MixPhaseParam = 1 !Jones' method as default
  QsplitRainParamA = 0.15_kind_real
  QsplitRainParamB = 0.09_kind_real
  QsplitRainParamC = 50.0_kind_real

  ! Compute saturated water vapor profile for nlevels_q only

  call Ops_Qsat (qsaturated(:),    & ! out
    t(:),             & ! in
    p(:), & ! in
    size(qtotal))                    ! in

  SmallValue = one / 8.5_kind_real
  Denom = SmallValue * (upper_rh - lower_rh)

  ! don't let rh exceed 2.0 to avoid cosh function blowing up
  RH_qtotal(:) = min (qtotal(:) / qsaturated(:), two)

  V1(:) = (RH_qtotal(:) - lower_rh) / Denom
  V2(:) = (RH_qtotal(:) - upper_rh) / Denom

  Y1 = one
  Y2 = Split_Factor
  Y3 = zero
  Y4 = Split_Factor

  Aconst = (Y2 - Y1) / two
  Bconst = -(Y4 - Y3) / two
  Cconst = (Y2 + Y1) / two
  Dconst = -(Y4 + Y3) / two

  ! Compute fraction of ql to ql+qi based on temperature profile

  where (t(:) - ZeroDegC >= -0.01_kind_real) ! -0.01degc and above

    ! all ql
    LF(:) = one

  end where

  where (t(:) <= MinTempql)

    ! all qi
    LF(:) = zero

  end where

  where (t(:) > MinTempql .and. t(:) - ZeroDegC < -0.01_kind_real)

    ! Jones' parametrization
    LF(:) = sqrt (-1.0_kind_real * log (-0.025_kind_real * (t(:) - ZeroDegC)) / 70.0_kind_real)

  end where

  ! finally set LF to 0.0_kind_real for the rttov levels on which clw jacobians are not
  ! calculated since nlevels_mwclw < nlevels_q

  V1zero = -1.0_kind_real * lower_rh / Denom
  V2zero = -1.0_kind_real * upper_rh / Denom
  IntConst = -(Aconst * Denom * log (cosh (V1zero)) + Bconst * Denom * log (cosh (V2zero)))
  W(:) = Aconst * Denom * log (cosh (V1(:))) + Bconst * Denom * log (cosh (V2(:))) + &
    (Cconst + Dconst) * RH_qtotal(:) + IntConst

  ! store the components of qtotal
  ! ensuring that they are above lower limits

  if (UseQtsplitRain) then

    ! Split qtotal into q and qnv (non-vapour part - includes
    ! ql, qi, qr)

    ! Split qtotal into q, ql ,qi

    do i = 1, size(qtotal)

      q(i) = max (W(i) * qsaturated(i), min_q)
      qnv(i) = max (qtotal(i) - q(i), zero)

      ! Split qnv into a cloud and precipitation part

      qc(i) = max (QsplitRainParamA * (QsplitRainParamB - (QsplitRainParamB / ((QsplitRainParamC * qnv(i)) + 1))), zero)

      ! Finally split non-precip part into liquid and ice

      ql(i) = max (LF(i) * qc(i), zero)
      qi(i) = max ((one - LF(i)) * (qc(i)), zero)

    end do

  else
    do i = 1, size(qtotal)

      q(i) = max (W(i) * qsaturated(i), min_q)
      ql(i) = max (LF(i) * (qtotal(i) - q(i)), zero)
      qi(i) = max ((one - LF(i)) * (qtotal(i) - q(i)), zero)

    end do

  end if

  ! Values of q, ql and qi are overwritten if output_type /= 1
  ! and replaced with the derivatives

  if (output_type /= 1) then

    ! Compute derivates
    ! q = dq/dqtotal, ql = dql/dqtotal, qi=dqi/dqtotal

    q(:) = Aconst * tanh (V1(:)) + Cconst + Bconst * tanh (V2(:)) + Dconst

    if (UseQtsplitRain) then

      ql(:) = LF(:) * QsplitRainParamA * QsplitRainParamB * QsplitRainParamC * (one - q(:)) /  &
        ((QsplitRainParamC * qnv(:)) + one) ** 2
      qi(:) = (one - LF(:)) * QsplitRainParamA * QsplitRainParamB * QsplitRainParamC * (one - q(:)) / &
        ((QsplitRainParamC * qnv(:)) + one) ** 2

    else

      ql(:) = LF(:) * (one - q(:))
      qi(:) = (one - LF(:)) * (one - q(:))

    end if

  end if

end subroutine Ops_SatRad_Qsplit

!-------------------------------------------------------------------------------
!> Do cholesky decomposition
!!
!! \details Heritage: Ops_Cholesky.inc
!!
!! Solves the Linear equation UQ=V for Q where U is a symmetric positive definite
!! matrix and U and Q are vectors of length N.  The method follows that in Golub
!! and Van Loan although this is pretty standard.
!!
!! if U is not positive definite this will be detected by the program and flagged
!! as an error.  U is assumed to be symmetric as only the upper triangle is in
!! fact used.
!!
!! \author Met Office
!!
!! \date 09/06/2020: Created
!!
subroutine Ops_Cholesky (U,         &
                         V,         &
                         N,         &
                         Q,         &
                         ErrorCode)

implicit none

! subroutine arguments:
integer, intent(in)          :: n
real(kind_real), intent(in)  :: U(n,n)
real(kind_real), intent(in)  :: V(n)
real(kind_real), intent(out) :: Q(n)
integer, intent(out)         :: ErrorCode

! Local declarations:
character(len=*), parameter  :: RoutineName = "Ops_Cholesky"
real(kind_real), parameter   :: Tolerance = tiny (zero) * 100.0_kind_real
character(len=50)            :: ErrorMessage
integer                      :: j
integer                      :: k
real(kind_real)              :: G(n,n)   ! The Cholesky Triangle Matrix
real(kind_real)              :: X(n)     ! Temporary array used in calculating G

ErrorCode = 0

! Determine the Cholesky triangle matrix.

do j = 1, n
  X(j:n) = U(j:n,j)
  if (j /= 1) then
    do k = 1, j - 1
      X(j:n) = X(j:n) - G(j,k) * G(j:n,k)
    end do
  end if
  if (X(j) <= Tolerance) then
    ErrorCode = 1
    Errormessage = ' :U matrix is not positive definite'
    write(*,*) trim(RoutineName), trim(ErrorMessage)
    goto 9999
  end if
  G(J:N,J) = X(J:N) / sqrt (X(J))
end do

! Solve Gx=v for x by forward substitution

X = V
X(1) = X(1) / G(1,1)
do j = 2, n
  X(j) = (X(j) - dot_product(G(j,1:j - 1), X(1:j - 1))) / G(j,j)
end do

! Solve G^T.q=x for q by backward substitution

Q = x
Q(n) = Q(n) / G(n,n)
do j = n - 1, 1, -1
  Q(j) = (Q(j) - dot_product(G(j + 1:n,j), Q(j + 1:n))) / G(j,j)
end do

9999 continue

end subroutine Ops_Cholesky

!-------------------------------------------------------------------------------
!> Find a free file unit.
!!
!! \author Met Office
!!
!! \date 09/06/2020: Created
!!
function ufo_utils_iogetfreeunit() result(unit)

implicit none

integer            :: unit

integer, parameter :: unit_min=10
integer, parameter :: unit_max=1000
logical            :: opened
integer            :: lun
integer            :: newunit

newunit=-1
do lun=unit_min,unit_max
  inquire(unit=lun,opened=opened)
  if (.not. opened) then
      newunit=lun
    exit
  end if
end do
unit=newunit

end function ufo_utils_iogetfreeunit

!-------------------------------------------------------------------------------
! Invert a matrix.
!
! Variables with intent in:
!
!   N:  Size of the matrix being inverted
!   M:  If MATRIX is not present this is the same as N, else this is the other
!       dimension of MATRIX.
!
! Variables with intent inout:
!
!   A:  Real matrix (assumed square and symmetrical) overwritten by its
!       inverse if MATRIX is not present.
!
! Variables with intent out:
!
!   ExitCode: 0: ok, 1: A is not positive definite.
!
! Optional variables with intent inout:
!
!   Matrix: If present this input matrix is replaced by (Matrix).A^-1 on exit
!           (leaving A unchanged).
!
! Uses Cholesky decomposition - a method particularly suitable for real
! symmetric matrices.  Cholesky decomposition solves the Linear equation UQ=V
! for Q where U is a symmetric positive definite matrix and U and Q are vectors
! of length N.  The method follows that in Golub and Van Loan although this is
!  pretty standard.  If U is not positive definite this will be detected by the
! program and flagged as an error.  U is assumed to be symmetric as only the
! upper triangle is in fact used.  If the the optional parameter Matrix is
! present, it is replaced by (Matrix).A^-1 on exit and A is left unchanged.
!-------------------------------------------------------------------------------

subroutine InvertMatrix (n,      &
                         m,      &
                         a,      &
                         status, &
                         matrix)

implicit none

! subroutine arguments:
integer, intent(in)                           :: n           !< order of a
integer, intent(in)                           :: m           !< order of optional matrix, if required
real(kind=kind_real), intent(inout)           :: a(n,n)      !< square mx, overwritten by its inverse
integer, intent(out)                          :: status      !< 0 if all ok 1 if matrix is not positive definite
real(kind=kind_real), optional, intent(inout) :: matrix(n,m) !< replaced by (matrix).a^-1 on exit

! local declarations:
character(len=*), parameter   :: routinename = "InvertMatrix"
character(len=80)             :: errormessage
real(kind=kind_real), parameter :: tolerance = tiny (zero) * 100.0_kind_real
integer                       :: i
integer                       :: j
integer                       :: k
integer                       :: mm
real(kind=kind_real)          :: g(n,n)      ! the cholesky triangle matrix
real(kind=kind_real)          :: q(n)
real(kind=kind_real)          :: tmp(n,m)
real(kind=kind_real)          :: v(n)
real(kind=kind_real)          :: x(n)        ! temporary array

status = 0

! determine the cholesky triangle matrix.

do j = 1, n
  x(j:n) = a(j:n,j)
  if (j /= 1) then
    do k = 1, j - 1
      x(j:n) = x(j:n) - g(j,k) * g(j:n,k)
    end do
  end if
  if (x(j) <= tolerance) then
    errormessage = 'Matrix is not positive definite'
    call fckit_log % warning(errormessage)
    status = 1
    goto 9999
  end if
  g(j:n,j) = x(j:n) / sqrt (x(j))
end do

! now solve the equation g.g^t.q=v for the set of
! vectors, v, with one element = 1 and the rest zero.
! the solutions q are brought together at the end to form
! the inverse of g.g^t (i.e., the inverse of a).

if (present (matrix)) then
   mm = m
else
  ! make sure that the dimensions of tmp were correctly specified
  if (m /= n) then
    errormessage = '2nd and 3rd arguments of routine must be'
    call fckit_log % warning(errormessage)
    errormessage = 'identical if the matrix option is not present'
    call fckit_log % warning(errormessage)
    status = 2
    goto 9999
  end if
  mm = n
end if

main_loop : do i = 1, mm
  if (.not. (present (matrix))) then
    v(:) = zero
    v(i) = one
  else
    v(:) = matrix(i,:)
  end if

  ! solve gx=v for x by forward substitution

  x = v
  x(1) = x(1) / g(1,1)
  do j = 2, n
    x(j) = (x(j) - dot_product (g(j,1:j - 1), x(1:j - 1))) / g(j,j)
  end do

  ! solve g^t.q=x for q by backward substitution

  q = x
  q(n) = q(n) / g(n,n)
  do j = n - 1, 1, -1
    q(j) = (q(j) - dot_product (g(j + 1:n,j), q(j + 1:n))) / g(j,j)
  end do

  tmp(:,i) = q(:)

end do main_loop

if (.not. (present (matrix))) then
  a(:,:) = tmp(:,:)
else
  matrix(:,:) = tmp(:,:)
end if

9999 continue

end subroutine InvertMatrix

!-------------------------------------------------------------------------------
! This function will comapre two strings while avoiding trim in order to speed
! up the string comparison. This will allow for trailing blanks on either string
! to count for a match.
!
! Since the strings may be different lengths coming in (due to trailing blanks)
! a helper function (cmp_ordered_strings) is called with the shorter string first
! which helps simplify the logic of the comparison.

function cmp_strings(str1, str2)
  implicit none

  logical :: cmp_strings
  character(len=*) :: str1
  character(len=*) :: str2

  if (len(str1) <= len(str2)) then
    cmp_strings = cmp_ordered_strings(str1, str2)
  else
    cmp_strings = cmp_ordered_strings(str2, str1)
  endif

  return
end function cmp_strings

!-------------------------------------------------------------------------------
! This is a helper function for the public cmp_strings function. It is intended
! for this function to be called with the arguments of cmp_strings, with the shorter
! of the two strings in the first argument.
!
! The algorithm is to compare each character one by one between each string. Then if
! one string is longer, continue looking at that string an make sure the remainder
! consists of blanks before calling it a match between the strings. Knowing that the
! first argument is the shorter of the two stings helps simplify the code doing the check.
!
! The trim call is avoided since it allocates, copies strings and deallocates which
! collectively are operations that are too expensive.

function cmp_ordered_strings(shorter_str, longer_str)
  implicit none

  logical :: cmp_ordered_strings
  character(len=*) :: shorter_str
  character(len=*) :: longer_str

  integer :: i
  integer :: j

  ! Check each character from the beginning of the strings to the length of the
  ! shorter string. Exit the loop right away if a mis-match occurs to save time.
  cmp_ordered_strings = .true.
  do i = 1, len(shorter_str)
    if (shorter_str(i:i) /= longer_str(i:i)) then
      cmp_ordered_strings = .false.
      exit
    endif
  enddo

  ! If cmp_ordered_strings is false we can return immediately. If true, then we need to check
  ! that the remainder of the longer string constists of all blank spaces before
  ! declaring the strings equal.
  if (cmp_ordered_strings) then
    do j = i, len(longer_str)
      if (longer_str(j:j) /= " ") then
        cmp_ordered_strings = .false.
        exit
      endif
    enddo
  endif

  return
end function cmp_ordered_strings

!------------------------------------------------------------------------------
!> Find the unique entries in the input list
!!
!! \author Met Office
!!
!! \date 15/10/2020: Created
!!
subroutine find_unique(input, output)

use, intrinsic :: iso_c_binding

integer(c_size_t), intent(in) :: input(:)
integer, allocatable, intent(out) :: output(:)

integer, allocatable :: unique_vals(:)    ! The list of unique elements
integer :: nfound                         ! The number of unique elements found
integer :: cur_val                        ! The current value being considered
integer :: max_val                        ! The maximum value in the input list

nfound = 0
allocate(unique_vals(1:size(input)))

cur_val = minval(input) - 1
max_val = maxval(input)
do while (cur_val < max_val)
    nfound = nfound + 1
    cur_val = minval(input, mask=input>cur_val)
    unique_vals(nfound) = cur_val
end do
allocate(output(nfound))
output = unique_vals(1:nfound)

end subroutine find_unique

!------------------------------------------------------------------------------
!> Generates a index array pointing to the elements of the array 'key'
!  in increasing order
!!
!! \author Met Office
!!
!! \date 15/10/2020: Created
!!
!
! Method:
!   The heap sort invented by J.W.J.Williams is used.
!   A description of the method can be found in 'Numerical Recipes'
!   The group information array is used to allow easy sorting on several
!   parameters of different types. For details see the Parent Module
!   OpsMod_Sort
!
! Inputs:
!   key : An array of character strings, to be sorted
!
! Input/Output:
!   index : An integer array pointing to the sorted items.
!-------------------------------------------------------------------------------
SUBROUTINE Ops_RealSortQuick(key,   &
                             index)

IMPLICIT NONE

! Subroutine arguments:
REAL(kind_real), INTENT(IN)       :: key(:)
INTEGER, ALLOCATABLE, INTENT(OUT) :: index(:)

! Local declarations:
INTEGER                           :: n     ! The number of items
INTEGER                           :: head  ! heaps are tree structures: head and child refer
INTEGER                           :: child ! to related items within the tree
INTEGER                           :: j
INTEGER                           :: dum   ! used to swap index items
CHARACTER(len=*), PARAMETER       :: RoutineName = 'Ops_RealSortQuick'

! Could put in an optional mask

n = SIZE (key)
ALLOCATE (Index(n))
DO j = 1, n
  Index(j) = j
END DO

! Do heapsort: Create the heap...

makeheap : DO j = n / 2, 1, -1
  head = j
  sift1 : DO

    ! find the largest out of the head and its two children...

    child = head * 2
    IF (child > n) EXIT sift1
    IF (child < n) THEN
      IF (key(Index(child + 1)) > key(Index(child))) child = child + 1
    END IF

    ! if the head is the largest, then sift is done...

    IF (key(Index(head)) >= key(Index(child))) EXIT sift1

    ! otherwise swap to put the largest child at the head,
    ! and prepare to repeat the procedure for the head in its new
    ! subordinate position.

    dum = Index(child)
    Index(child) = Index(head)
    Index(head) = dum
    head = child
  END DO sift1
END DO makeheap

! Retire heads of the heap, which are the largest, and
! stack them at the end of the array.

retire : DO j = n, 2, -1
  dum = Index(1)
  Index(1) = Index(j)
  Index(j) = dum
  head = 1

  ! second sift is similar to first...

  sift2: DO
    child = head * 2
    IF (child > (j - 1)) EXIT sift2
    IF (child < (j - 1)) THEN
      IF (key(Index(child + 1)) > key(Index(child))) child = child + 1
    END IF
    IF (key(Index(head)) >= key(Index(child))) EXIT sift2
    dum = Index(child)
    Index(child) = Index(head)
    Index(head) = dum
    head = child
  END DO sift2
END DO retire

END SUBROUTINE Ops_RealSortQuick
!-------------------------------------------------------------------------------

subroutine sort_and_unique(nobs, impact_param, record_number, index_vals, unique)

use, intrinsic :: iso_c_binding

integer, intent(in)               :: nobs                  ! Total number of observations
real(kind_real), intent(inout)    :: impact_param(1:nobs)  ! The impact parameter for GNSS-RO
integer(c_size_t), intent(inout)  :: record_number(1:nobs) ! Number used to identify unique profiles in the data
integer, allocatable, intent(out) :: index_vals(:)         ! Indices of sorted observation
integer, allocatable, intent(out) :: unique(:)             ! Set of unique profile numbers

real(kind_real), allocatable  :: sort_key(:)           ! Key for the sorting (based on record number and impact parameter)
integer                       :: start_point           ! Starting index of the current profile
integer                       :: current_point         ! Ending index of the current profile
integer                       :: iprofile              ! Loop variable, profile number

! Read through the record numbers in order to find a profile of observations
! Each profile shares the same record number
allocate(sort_key(nobs))
sort_key = record_number + (impact_param / MAXVAL(impact_param))
call Ops_RealSortQuick(sort_key, index_vals)
call find_unique(record_number, unique)

end subroutine sort_and_unique

  FUNCTION upper2lower(str) RESULT(string)

    IMPLICIT NONE

    CHARACTER(*), INTENT(in) :: str
    CHARACTER(LEN(str))      :: string

    INTEGER :: ic, i

    CHARACTER(26), PARAMETER :: upper = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    CHARACTER(26), PARAMETER :: lower = 'abcdefghijklmnopqrstuvwxyz'

!   lowcase each letter if it is lowecase
    string = str
    DO i = 1, LEN_TRIM(str)
      ic = INDEX(upper, str(i:i))
      IF (ic > 0) string(i:i) = lower(ic:ic)
    END DO

  END FUNCTION upper2lower

  INTEGER FUNCTION getindex(names,usrname)
    IMPLICIT NONE
    CHARACTER(len=*),INTENT(in) :: names(:)
    CHARACTER(len=*),INTENT(in) :: usrname
    INTEGER i
    getindex=-1
    DO i=1,SIZE(names)
      IF(TRIM(usrname)==TRIM(names(i))) THEN
        getindex=i
        EXIT
      ENDIF
    ENDDO
  END FUNCTION getindex

end module ufo_utils_mod
